/* 
  Implementar un algoritmo de sincronizacion para el transporte de entidades
  de dos tipos diferentes. Se deben cumplir las siguientes restricciones:
  1. Existe un medio de transporte que tiene un cupo (par) determinado.
  2. Debido a problemas entre las entidades, se debe transportar
     cupo/2 entidades de cada tipo en cada viaje.
  3. Suponga que la cantidad de entidades de los dos tipos que desean viajar
     es valida, esto quiere decir que siempre se pueden trasportar 
     en cantidades de cupo/2 (siempre se llena el medio de transporte).
*/

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <time.h>
#include <semaphore.h>

#define N 50
#define M 50
#define TOTAL N + M
#define CUPO 10

typedef void (*thread_body)(void * arg);
typedef sem_t semaphore;

typedef struct {
  pthread_t id;
  pthread_attr_t attr;
  thread_body start_func;
  void * arg;
}thread;

typedef struct {
	int id;
	int type;
}thread_parameter;

semaphore mutex;
semaphore mutex_cero;
semaphore mutex_uno;
semaphore subida_cero;
semaphore subida_uno;
semaphore viaje;
semaphore mutex_bajada;
int cont_cero;
int cont_uno;


thread * start_thread(thread_body func, void * arg) {
  thread* t;
  t=(thread*)malloc(sizeof(thread));
  t->start_func=func;
  t->arg = arg;
  pthread_attr_init(&(t->attr));
  pthread_create(&(t->id), &(t->attr), (void*)t->start_func, t->arg);
  return t;
}

void down(semaphore *sem) {
  sem_wait(sem);
}

void up(semaphore *sem) {
  sem_post(sem);
}

void passenger(void * param) {
	thread_parameter * tp;
	int id;
	int type;
	int i;

	tp = (thread_parameter *) param;

	id = tp->id;
	type = tp->type;
	
	

	printf("Thread %d (%d) started\n", id, type);
	sleep(rand() % 3); //Simular una demora aleatoria en la llegada
	//Llegada del proceso
	if (type == 0) {
		down(&mutex_cero);
	}else {
		down(&mutex_uno);
	}
	
	down(&mutex); //Seccion critica: tratar de subirse al transporte
	    if (type == 0) {
		cont_cero = cont_cero + 1;
		if (cont_cero == (CUPO / 2)) {
		   for (i=0; i<(CUPO/2); i++) {//Autorizar la subida
			up(&subida_cero);
		   }
		}else {
		   up(&mutex_cero); //Dar paso al siguiente del tipo 0
	        }
	    }else {
		cont_uno = cont_uno + 1;
		if (cont_uno == (CUPO / 2)) { 
		   for (i=0; i<(CUPO/2); i++) { //Autorizar la subida
			up(&subida_uno);
		   }
		}else {
		   up(&mutex_uno); //Dar paso al siguiente del tipo 1
	        }
	    }
	    if (cont_cero + cont_uno == CUPO) { //Se completo el viaje
		for (i=0; i<CUPO; i++) {
			up(&viaje);
		}
	    }else {
		up(&mutex);
	    }		
	
	//printf("\tThread %d (%d) waiting transport\n", id, type);
	if (type == 0) {
		down(&subida_cero);
	}else {
		down(&subida_uno);
	}
	
	printf("\tThread %d (%d) on transport!\n", id, type);

	down(&viaje);
	
	printf("\t\tThread %d (%d) on traveling...\n", id, type);
	sleep(1);

	down(&mutex_bajada); //Seccion Critica

	if (type == 0) {
		cont_cero = cont_cero - 1;
	}else {
		cont_uno = cont_uno - 1;
	}

	
	printf("\t\t\tThread %d (%d) getting down..\n", id, type);
	if (cont_cero + cont_uno == 0) { //Ultimo en bajarse?
		printf("\t\t\tThread %d (%d) is the last passenger!\n", id, type);
		up(&mutex_cero);
		up(&mutex_uno);
		up(&mutex);
	}

	up(&mutex_bajada);
}


int ids[N];
thread * threads[N+M];

int main(int argc, char * argv[]) {
  int i;
  int n;
	
  thread_parameter *tp;


  printf("Main Started\n");

  sem_init(&mutex, 0, 1);
  sem_init(&mutex_cero, 0, 1);
  sem_init(&mutex_uno, 0, 1);	
  sem_init(&mutex_bajada, 0, 1);

  sem_init(&subida_cero, 0, 0);
  sem_init(&subida_uno, 0, 0);	

  sem_init(&viaje, 0, 0);

  cont_cero = 0;
  cont_uno = 0;

  n = 1; 
  for (i=0; i<TOTAL; i++) {  
      tp = (thread_parameter *)malloc(sizeof (thread_parameter));
      tp->id=n;
      if (i %2 == 0) {
	tp->type = 0;
        threads[i] = start_thread(passenger, (void*)tp);
      }else {
        tp->type = 1;
        threads[i] = start_thread(passenger, (void*)tp);
      }
      n = n + 1;
  }

  for (i=0; i<TOTAL; i++) {
	pthread_join(threads[i]->id, NULL);
  }  

  printf("Main Finished\n");
  exit(0);
}
